
"""
The hard filter class takes in a detector and returns
the set of errors that align with mispredictions (a set of indices)
"""
class HardFilter:

	def __init__(self, 
				 detector,
				 mispredictions):

		self.detector = detector
		self.mispredictions = set(mispredictions)

	def fit(self):
		self.detector.fit()
		self.detector.error_list = [v for v in self.detector.error_list if v[0] in self.mispredictions]
		self.detector.iterator = self.detector.error_list.__iter__()

	#iterator interface to the error detector
	def __iter__(self):
		return self

	def next(self):
		return self.detector.next()




import numpy as np
import copy
import sklearn
from sklearn.cross_validation import KFold
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import classification_report
from preprocessing_utils import *

"""
This module uses a trained model to identify
possibly dirty records
"""

#uses a decision tree to characterize the mispredictions
def generate_constraints(X, y, misp, 
						   max_depth=4, 
						   min_samples_leaf=10,
						   slack = 0.1,
						   detection_rate=0.9):

	c = DecisionTreeClassifier(max_depth=max_depth, 
							   min_samples_leaf=min_samples_leaf, 
							   criterion='entropy')

	labels = np.zeros((X.shape[0],1))

	labels[misp] = 1
	c.fit(X,labels)
	y_pred = c.predict(X)

	paths = get_decision_paths(c)

	return (c, get_cleaning_rules(paths, slack), get_detection_rules(paths, detection_rate))

#using the tree this gets all of the decision paths
def get_decision_paths(tree):
	left      = tree.tree_.children_left
	right     = tree.tree_.children_right
	threshold = tree.tree_.threshold
	value = tree.tree_.value

	allpaths = []

	def recurse(left, right, threshold, node, lset=[]):

		if (threshold[node] != -2):
			if left[node] != -1:
				nlset = copy.copy(lset)
				nlset.append((tree.tree_.feature[node], 1 , threshold[node]))
				recurse (left, right, threshold, left[node], nlset)
			if right[node] != -1:
				nlset = copy.copy(lset)
				nlset.append((tree.tree_.feature[node], -1 , threshold[node]))
				recurse (left, right, threshold, right[node], nlset)
		else:
			nlset = copy.copy(lset)
			nlset.append((tree.tree_.feature[node], threshold[node], np.squeeze(value[node])))
			allpaths.append(nlset)

	recurse(left, right, threshold, 0)

	return allpaths


#this generates a set of contraints given the tree
def get_cleaning_rules(paths, slack):

	#get all paths that have more than 90 accuracy
	clean_paths = [p for p in paths if p[len(p)-1][2][1]/np.sum(p[len(p)-1][2]) < slack]

	total = np.sum(np.array([p[len(p)-1][2] for p in paths]).flatten())
	covered_total = np.sum(np.array([p[len(p)-1][2] for p in clean_paths]).flatten())

	return (clean_paths, covered_total/total)


#this generates a set of detection rules
def get_detection_rules(paths, accuracy):

	#get all paths that have more than 90 accuracy
	dirty_paths = [p for p in paths if p[len(p)-1][2][1]/np.sum(p[len(p)-1][2]) > accuracy]

	total = np.sum(np.array([p[len(p)-1][2] for p in paths]).flatten())
	covered_total = np.sum(np.array([p[len(p)-1][2] for p in dirty_paths]).flatten())

	return (dirty_paths, covered_total/total)

#get all safe records
def get_all_safe_records(X, constraints):
	
	safe_records = []

	for i in range(X.shape[0]):

		
		for constraint in constraints:
			
			safe = True
			
			vals = []

			for node in constraint[0:len(constraint)-1]:

				#vals.append(X[i,node[0]])

				if node[1] == 1 and X[i,node[0]] > node[2]:
					safe = False
					break

				if node[1] == -1 and X[i,node[0]] <= node[2]:
					safe = False
					break

			if safe:
				safe_records.append(i)
				break
	
	unsafe_records = set(range(X.shape[0])).difference(set(safe_records))

	return np.array(safe_records), np.array(list(unsafe_records))


#get all the violated constraints
def get_violations(X, unsafe, constraints):

	constraint_violations = []

	for i in unsafe:

		vals = []

		for constraint in constraints:

			node_vals = []

			for node in constraint[0:len(constraint)-1]:

				if node[1] == 1 and X[i,node[0]] > node[2]:
					
					node_vals.append((constraint, node[1], i, X[i,node[0]], node[0], node[2]))

				if node[1] == -1 and X[i,node[0]] <= node[2]:

					node_vals.append((constraint, node[1], i, X[i,node[0]], node[0], node[2]))

			vals.append(node_vals)

		constraint_violations.append((i,vals))	

	return constraint_violations

#sample edits to rectify the constraint violations
def get_safeset_violations( dataset, 
							types,
							true_labels,
							mispredictions,
							slack=0.35, 
							max_depth=4, 
						    min_samples_leaf=10):

	X = featurize(dataset, types)
	
	constraints = generate_constraints(X, true_labels, mispredictions, slack=slack, 
									   max_depth=max_depth, 
									   min_samples_leaf=min_samples_leaf)

	return get_all_safe_records(X, constraints[1][0])[1]










"""
Error Detectors can take as input a Logging
class
"""
from Logging import Logging
import csv
import datetime
import json

class CSVLogging(Logging):
	
	def __init__(self,filename):
		self.fd = open(filename, 'a')
		self.csvw = csv.writer(self.fd)

		self.ERROR = 0
		self.SCHEMA = 1
		self.CONFIG = 2
		self.TYPE = 3


	def logError(self, error_object):
		self.csvw.writerow([self.ERROR, 
							datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
							json.dumps(error_object)])

	def logFileType(self, size, cols):
		self.csvw.writerow([self.TYPE, 
							datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
							json.dumps([size, cols])])

	def logSchema(self, type_array):
		self.csvw.writerow([self.SCHEMA, 
							datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
							json.dumps(type_array)])

	def logConfig(self, module_array, config_array):
		self.csvw.writerow([self.CONFIG, 
							datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
							json.dumps((str(module_array), config_array))])



"""
Detects Distribution Errors in a discrete distribution
dataset using standard deviations
"""
import numpy as np
from ErrorModule import ErrorModule

class DistributionErrorModule(ErrorModule):

	def __init__(self, thresh=3.5, fail_thresh=2):
		self.thresh = thresh
		self.fail_thresh = fail_thresh

	"""
	Returns the subset of a domain that is potentially
	erroneous
	"""
	def predict(self, vals):

		dist = {}

		for v in vals:
			if v not in dist:
				dist[v] = 0

			dist[v] = dist[v] + 1

		
		valsv = set(vals)
		valsd = [dist[v] for v in valsv]


		mad = np.std(valsd)

		median = np.mean(valsd)
		
		#make a copy
		incorpus = []
		error = []
		incorpus.extend(valsv)


		#fail if less than the fail threshold, or if
		#mad not informative
		if len(valsd) <= self.fail_thresh or mad < 1.0:
			return error, incorpus

		for a in valsv:
			
			if np.abs(dist[a] - median) > self.thresh*mad:
				#print a, dist[a], mad, median
				error.append(a)
				incorpus.remove(a)

		return error, incorpus

		
	"""
	Turns an error set into a set of records
	"""
	def getRecordSet(self, errors, dataset, col):

		indices = []
		erecords = []

		for i,d in enumerate(dataset):
			val = d[col]

			match = False

			for e in errors:

				if e == val:
					match = True

			if match:
				indices.append(i)
				erecords.append(d)

		return erecords, indices

	"""
	Returns a description
	"""
	def desc(self):
		return "A distinct value was found with a count of greater than > " + str(self.thresh) + " stds from the mean count"


	def availTypes(self):
		return ['numerical', 'categorical', 'string']


import numpy as np
import usaddress

class LoLTypeInference:
	"""
	Given a list of lists this class infers the types of
	each of the attributes.
	"""

	def __init__(self, cat_thresh=0.25, 
					   num_thresh = 0.25, 
					   addr_thresh=0.25):
		"""
		cat_thresh parsing threshold for categorical attrs
		num_thresh parsing threshold for numerical attributes
		"""

		self.cat_thresh = cat_thresh
		self.num_thresh = num_thresh
		self.addr_thresh = addr_thresh


	def getDataTypes(self, data):
		"""
		Given data in a list of lists format this returns a list 
		of attribute types
		"""

		num_attributes = len(data[0])

		type_array = []

		for i in range(0, num_attributes):

			if self.__is_addr(data, i):
				type_array.append('address')
			elif self.__is_num(data, i):
				type_array.append('numerical')
			elif self.__is_cat(data, i):
				type_array.append('categorical')
			else:
				type_array.append('string')

		return type_array


	def __is_num(self, data, col):
		"""
		Internal method to determine whether data is numerical
		"""
		float_count = 0.0
		for datum in data:
			try:
				float(datum[col].strip())
				float_count = float_count + 1.0
			except ValueError:
				pass

		return (float_count/len(data) > self.num_thresh)

	def __is_cat(self, data, col):
		"""
		Internal method to determine whether data is categorical
		defaults to number of distinct values is N/LogN
		"""
		counts = {}

		for datum in data:
			d = datum[col]

			if d not in counts:
				counts[d] = 0

			counts[d] = counts[d] + 1

		total = len([k for k in counts if counts[k] > 1])+0.0

		return (total/len(counts) > self.cat_thresh)


	def __is_addr(self, data, col):
		"""
		Internal method to determine whether the column is an address
		"""
		addr_count = 0.0
		for datum in data:
			try:
				vals = usaddress.parse(datum)
				addrstrings = [v for v in vals if v[1] == 'Recipient']
				if len(addrstrings) != len(vals):
					addr_count = addr_count + 1.0
			except:
				pass

		return (addr_count/len(data) > self.addr_thresh)




"""
Error Detectors can take as input a Logging
class
"""

class Logging:
	"""
	defines the abstract class for logging
	"""

	def logError(self, error_object):
		"""
		Given a dict generated by an error dector, logs
		"""
		raise NotImplemented("Loggers must implement logError")


	def logSchema(self, type_array):
		"""
		Given a dict generated by the type inference
		"""
		raise NotImplemented("Loggers must implement logSchema")

	def logFileType(self, size, cols):
		"""
		Given a dict generated by the type inference
		"""
		raise NotImplemented("Loggers must implement logFileType")


	def logConfig(self, module_array, config_array):
		"""
		Given a dict input to the detector
		"""
		raise NotImplemented("Loggers must implement logConfig")



"""
An error module is one of the basic building blocks of ActiveDetect.
This error detector applies to the distinct values of a column.
It provides four methods predict(domain), getRecordSet(error_domain), availTypes(), desc()
"""

class ErrorModule:


	def predict(self, vals):
		"""
		The return value is expected to be two sets error, not error
		"""
		raise NotImplemented("An error module must implement predict")

	def getRecordSet(self, errors, dataset, col):
		"""
		The return value is a list of erroneous records, their indices
		"""
		raise NotImplemented("An error module must implement getRecordSet")

	def desc(self):
		"""
		All error modules should have a human readable description of what the error is
		"""
		raise NotImplemented("An error module must implement desc")

	def availTypes(self):
		"""
		Returns a data types that are supported
		"""
		raise NotImplemented("An error module must implement availTypes")


from setuptools import setup, find_packages

setup(
  name = 'activedetect',
  packages = find_packages(), # this must be the same as the name above
  version = '0.1.4-2',
  description = 'A Library For Error Detection For Predictive Analytics',
  author = 'Sanjay Krishnan and Eugene Wu',
  author_email = 'sanjay@eecs.berkeley.edu',
  url = 'https://github.com/sjyk/activedetect/', # use the URL to the github repo
  download_url = 'https://github.com/sjyk/activedetect/tarball/0.1.4-2', 
  keywords = ['error', 'detection', 'cleaning'], # arbitrary keywords
  classifiers = [],
  install_requires=[
          'numpy',
          'sklearn',
          'gensim',
          'usaddress',
          'scipy'
      ]
)




from preprocessing_utils import *
from model_eval_utils import *
from activedetect.loaders.type_inference import LoLTypeInference


"""
The safe set filter class takes in a detector and returns
the set of errors that align with mispredictions and are outliers--not just random variation in the data.
"""
class SafeSetFilter:

	def __init__(self, 
				 detector,
				 mispredictions,
				 unlabeled_dataset,
				 labels, 
				 #safe set config below
				 max_depth=4, 
				 min_samples_leaf=10,
				 slack = 0.1,
				 detection_rate=0.9):

		self.detector = detector
		self.mispredictions = mispredictions
		self.udataset = unlabeled_dataset
		self.labels = labels

		self.max_depth = max_depth
		self.min_samples_leaf = min_samples_leaf
		self.slack = slack
		self.detection_rate = detection_rate

	def fit(self):
		self.types = LoLTypeInference().getDataTypes(self.udataset)

		self.safeset = set(get_safeset_violations(self.udataset, 
												  self.types, 
												  self.labels, 
												  self.mispredictions,
												  max_depth=self.max_depth, 
				 								  min_samples_leaf=self.min_samples_leaf,
				 								  slack = self.slack,
				 								  detection_rate=self.detection_rate))

		self.detector.fit()
		self.detector.error_list = [v for v in self.detector.error_list if v[0] in self.mispredictions and v[0] in self.safeset]
		self.detector.iterator = self.detector.error_list.__iter__()

	#iterator interface to the error detector
	def __iter__(self):
		return self

	def next(self):
		return self.detector.next()




"""
Given a dataset returns the cells that 
are possibly erroneous.
"""
import numpy as np
from activedetect.loaders.type_inference import LoLTypeInference
from StringSimilarityErrorModule import StringSimilarityErrorModule
from QuantitativeErrorModule import QuantitativeErrorModule
from SemanticErrorModule import SemanticErrorModule
from DistributionErrorModule import DistributionErrorModule

class ErrorDetector:

	def __init__(self, 
				 dataset, 
				 cols = None,
				 modules = [],
				 config =  []):	

		if len(config) != len(modules):
			raise ValueError("Config must be the same length as the modules list")

		if len(modules) == 0:
			self.default__init__(dataset, cols)
			return

		self.modules = modules
		self.config = config

		self.cols = cols

		self.dataset = dataset

		self.types = LoLTypeInference().getDataTypes(dataset)

		self.modules = [d(**config[i]) for i, d in enumerate(modules)] 

		self.all_errors = [[[] for x in range(len(self.types))] for y in range(len(dataset))] 
		
		self.error_list = []

		self.iterator = None

		self.logger = None


	"""
	Adds a logger to the error detector
	"""
	def addLogger(self, logger):
		self.logger = logger

	"""
	default detectors and config
	"""
	def default__init__(self, dataset, cols):
		d_detect = DistributionErrorModule
		q_detect = QuantitativeErrorModule
		s_detect = SemanticErrorModule
		str_detect = StringSimilarityErrorModule
		config = [{'thresh':20}, 
				  {'thresh': 10}, 
				  {'thresh': 10, 'corpus': 'corpora/text8'}, 
				  {'thresh': 10}]

		return self.__init__(dataset, cols, [d_detect,q_detect, s_detect, str_detect], config)



	"""
	Based on the data type this applies one of the detectors
	"""
	def __predictCol(self, col):
		col_type = self.types[col]

		error_dict = {}
		for d in self.modules:
			if col_type in d.availTypes():
				vals = [r[col] for r in self.dataset]
				error_dict[d.desc()] = d.getRecordSet(d.predict(vals)[0], self.dataset, col)

		return error_dict

	"""
	The main user facing method to get all of the errors
	"""
	def fit(self):
		
		#log schema, config, file type
		if self.logger != None:
			self.logger.logSchema(self.types)
			self.logger.logFileType(len(self.dataset), len(self.types))
			self.logger.logConfig(self.modules, self.config)

		for i,t in enumerate(self.types):

			if  not self.cols == None and i not in self.cols:
				continue 

			output = self.__predictCol(i)

			for k in output:
				for index in output[k][1]:
					self.all_errors[index][i].append(k)
					self.error_list.append((index, i))
					
					if self.logger != None:
						self.logger.logError(self.errorToDict((index, i)))

		self.iterator = self.error_list.__iter__()
		
		return self.all_errors

	
	#iterator interface to the error detector
	def __iter__(self):
		return self

	def errorToDict(self, v):
		return {'cell': v, 
				'error_types':self.all_errors[v[0]][v[1]],
				'cell_value': self.dataset[v[0]][v[1]],
				'record_value': self.dataset[v[0]]}

	#gets the next error
	def next(self):
		v = self.iterator.next()
		return self.errorToDict(v)










